﻿'--------------------------------------------------------------------------------
'--        File: ExifToolIO.vb
'--
'-- Description: VB.Net library interface to Perl exiftool application script by Phil Harvey.
'--              See http://owl.phy.queensu.ca/~phil/exiftool/ for details
'--
'--     License: This is software, in whole or part, is free for use in
'--              non-commercial applications, provided that this copyright notice
'--              is retained.  A licensing fee may be required for use in a
'--              commercial application.
'--
'--     History: 
'-- 2014-06-09 - NCC - Added My.Application.Info.AssemblyName to temp exiftool.exe name to make it unique to this program.
'--                    Was having problem if a program was already running using ExifToolIO and then another different one
'--                    was started and DeleteTempProcesses = True, it would kill and delete the running ExifToolIO process, 
'--                    not good! So as it is now with DeleteTempProcesses = True, do not start another program that uses ExifToolIO
'--                    and has the same My.Application.Info.AssemblyName (program name)
'-- 2014-05-27 - NCC - Created,  curtis1757 at gmail dot com
'--------------------------------------------------------------------------------



Imports System.IO
Imports System.Text
Imports System.Threading
Imports System.ComponentModel
Imports System.Configuration
Imports System.Reflection
Imports System.Xml.Serialization

#Region "ExiftoolIO Options"

''' <summary>
''' These options may be changed from their default values to control how Exiftool is used.
''' </summary>
''' <remarks></remarks>
<XmlRoot(ExifToolIOOptions.ElementName), XmlType("", namespace:="")>
Public Class ExifToolIOOptions
    Public Const ElementName As String = "ExiftoolIOOptions"
    ''' <summary>
    ''' Sometimes if exiftool is not shutdown cleanly then temprary files and processes my remain active.  
    ''' If set to True (default) then these files and processes are looked for during Initialize and the 
    ''' processes are 'killed' and the files deleted.  
    ''' </summary>
    ''' <remarks>Initialize creates the exiftool processes with a unique name so they can be uniquely identified
    ''' and killed later if needed.</remarks>
    <XmlAttribute("DeleteTempProcesses")> Public DeleteTempProcesses As Boolean = True
    ''' <summary>
    ''' Set to True (default) to use stdinp for the input command file.  This is the fastest, but there will be no cmd
    ''' file to look at after execution for debugging purposes.  Set to False and provide a valid file name in 
    ''' CommandFileName for command file, whis will be a text file. Then if when Close is called and
    '''  the DeleteCommandFile parameter is set to False, the CommandFileName file will still be present to look
    ''' at for debugging purposes.
    ''' </summary>
    ''' <remarks></remarks>
    <XmlAttribute("UseStdInpCommandFile")> Public UseStdInpCommandFile As Boolean = True
    ''' <summary>
    ''' Provides the file name to use for the exiftool commands.  Only used if UseStdInpCommandFile is False.  Default
    ''' name of "-" is the same as setting UseStdInpCommandFile to true, otherwise provide a valid file name path.
    ''' Also stdinp will be used if the provided file name is "standardinput" or "stdinp".  If a blank file name is
    ''' given then a temporoary file name will be created and used.
    ''' </summary>
    ''' <remarks></remarks>
    <XmlAttribute("CommandFileName")> Public CommandFileName As String = "-"
    ''' <summary>
    ''' Provides the file name of the exiftool.exe file.  This is not used if UseResourcesExifToolExe is True. 
    ''' If an existing file name is given and UseResourcesExifToolExe is False, then this file will be used for the 
    ''' exiftool.exe file.  The Resources included in this .dll include a version of the exiftool.exe file, so it is not
    ''' required to have the exiftool.exe as a seperate file.  This file is likely not the most up to date version of 
    ''' exiftool, but it will work.  With the sources code to this .Net Project this filecan be updated.  
    ''' Use the ResourcesVersion property to determine what the exiftool.exe version is that is in the Resources.
    ''' </summary>
    ''' <remarks></remarks>
    <XmlAttribute("ExifToolExeFileName")> Public ExifToolExeFileName As String = Nothing
    ''' <summary>
    ''' If set to True (default) the the exiftool.exe file in Resources will be used.  Since it is required by WEindows to only
    ''' execute a process from a file, we will need to create a temporary file from the one in Resources.  This is created in
    ''' the users Local Application Data folder.
    ''' </summary>
    ''' <remarks>see CommandFileName summary for more info</remarks>
    <XmlAttribute("UseResourcesExifToolExe")> Public UseResourcesExifToolExe As Boolean = True
End Class
#End Region

''' <summary>
''' ExifToolIO Class provides Shared methods to Load exiftool.exe as a new process and keep the process open
''' interact with it.
''' Features:
''' -A version exiftool.exe is included in the ExifToolIO.dll so no seperate exiftool.exe is needed.  Optionally 
''' a seperate exiftool.exe file can be used (see Options above)
''' -I/O to the exiftool process is by default via stdinp and stdout so no disk file I/O is involved, 
''' making it very fast. Optionally a text file may be used and then viewed later to see what command were
''' actually sent to exiftool.
''' -Very simple to use, simply call Initiailize, send commands using Cmd method then call Execute which returns
''' the results of exiftool
''' </summary>
''' <remarks></remarks>
Public Class ExifToolIO

#Region "Private variables and constants"
    Private Const NTimeChars As Integer = 8

    Private Const ProcessExitWaitTime As Integer = 10000

    Private Const CmdQuiet = "-q"
    Private Const CmdTableFormat = "-T"
    Private Const CmdExecute = "-execute"
    Private Const CmdStayOpen = "-stay_open"
    Private Const ReadyString As String = "{ready}"

    Private Shared CmdFn As String
    Private Shared CmdStream As StreamWriter

    Private Shared ExifToolProcess As Process

    Private Shared stdoutString As New StringBuilder
    Private Shared stdoutReadyHandle As AutoResetEvent

    Private Shared erroutString As New StringBuilder
    Private Shared erroutWaitHandle As AutoResetEvent

    Private Shared CmdString As New StringBuilder

    Private Shared _LastErrorMsg As String

    Private Shared InitializationDone As Boolean = False

    Private Shared ReadOnly lockObject As New Object

    Private Shared CreatedTempExe As Boolean = True
    Private Shared ExifToolExeFn_ As String
    Private Shared ExifToolCmdFn_ As String

    Private Const StdInpFileName As String = "-"

    Private Shared _LoadedExifToolVer As String = "<ExifTool not initialized yet.>"
    Private Shared _ResourcesExifToolVer As String = My.Resources.ExifToolVersion
    Private Shared ExifToolProcessNameBase As String = "_exiftool" & My.Application.Info.AssemblyName

#End Region

#Region "Public Properties, Subs and Functions"

    Public Shared Options As New ExifToolIOOptions

    ''' <summary>
    ''' Returns the version of the exiftool.exe that is currently loaded. Which could be from either 
    ''' the My.Resources version or a user suppled version
    ''' </summary>
    ''' <value></value>
    ''' <returns>A string in the from ##.## that is the reported by the exiftool -ver command</returns>
    ''' <remarks>If exiftool fails to start proeprly for anyreason in Initialize, then this returns a 
    ''' message indicating why.</remarks>
    Public Shared ReadOnly Property LoadedVersion As String
        Get
            Return _LoadedExifToolVer
        End Get
    End Property

    ''' <summary>
    ''' Returns the version of the exiftool.exe that is in My.Resources
    ''' </summary>
    ''' <value></value>
    ''' <returns>A string in the from ##.## that is was 'hand' input when the exiftoolo.exe was added to My.Resources</returns>
    ''' <remarks>This needs to be updated in My.Resources whenever a new version of Exiftool.exe is put into My.Resources</remarks>
    Public Shared ReadOnly Property ResourcesVersion As String
        Get
            Return _ResourcesExifToolVer
        End Get
    End Property

    ''' <summary>
    ''' Initializes the exiftool IO using the options given in ExifToolIOOptions
    ''' </summary>
    ''' <returns>Tree if initialization succesfully completed and False otherwise</returns>
    ''' <remarks></remarks>
    Public Shared Function Initiailize() As Boolean
        If Not InitializationDone Then
            SyncLock lockObject

                '--If requested, try and remove any exiftool processes and temprary files left over from 
                '--previous failed Closings of exiftool. (if any)

                If Options.DeleteTempProcesses Then DoKillExifToolProcesses()

                Try
                    If Options.UseResourcesExifToolExe OrElse
                       String.IsNullOrEmpty(Options.ExifToolExeFileName.Trim) OrElse
                       Not File.Exists(Options.ExifToolExeFileName.Trim) Then

                        '--Get ExifTool.exe from Resources and copy as file to ApplicationData folder (Process needs .exe in a file)

                        Dim NowString As String = Now.Ticks.ToString

                        '--Create a unique name (using the current tick count) for the exiftool.exe file we are creating from Resources

                        ExifToolExeFn_ = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData) & Path.DirectorySeparatorChar &
                                        NowString.Substring(NowString.Length - NTimeChars) & ExifToolProcessNameBase & ".exe"

                        File.WriteAllBytes(ExifToolExeFn_, My.Resources.exiftool)
                        CreatedTempExe = True
                    Else

                        '--Use provided ExifTool executable file

                        ExifToolExeFn_ = Options.ExifToolExeFileName
                        CreatedTempExe = False
                    End If

                    If Options.UseStdInpCommandFile Then
                        ExifToolCmdFn_ = StdInpFileName
                    Else
                        ExifToolCmdFn_ = Options.CommandFileName
                    End If

                    '--Main routine to start up the exiftool process

                    StartExifToolProcess()

                    InitializationDone = True

                    _LoadedExifToolVer = Execute("-ver").Trim

                    If CreatedTempExe Then _ResourcesExifToolVer = _LoadedExifToolVer
                Catch ex As Exception

                    '--had some kind for failure

                    _LoadedExifToolVer = "<ExifTool failed to start: " & ex.Message & ">"
                End Try

            End SyncLock
        End If

        Return InitializationDone
    End Function

    ''' <summary>
    ''' Adds to the commands that will be used for the next exiftool execute done in Execute
    ''' </summary>
    ''' <param name="cmd">String: command string that will be added to the list</param>
    ''' <param name="ReplaceSpaceWithCRLF">Boolean: since each command needs to be on a seperate line, for
    ''' convience several commands can be given in a single 'cmd' seperated by a 'space' and those spaces
    ''' will be converted to CrLF so each will be on a seperate line.  Note that some commands can have
    ''' a space in them and for these commands this parameter should be set to False, default is True.</param>
    ''' <remarks></remarks>
    Public Shared Sub Cmd(cmd As String, Optional ReplaceSpaceWithCRLF As Boolean = True)
        If ReplaceSpaceWithCRLF Then
            CmdString.Append(cmd.Trim.Replace(" ", ControlChars.CrLf))
        Else
            CmdString.Append(cmd)
        End If

        CmdString.Append(ControlChars.CrLf)
    End Sub

    ''' <summary>
    ''' Executes the exiftool commands that have been 'loaded' via the Cmd sub
    ''' </summary>
    ''' <param name="fn">Optional String: Provides the file name that the exiftooo command is to act upon.  This is optional
    ''' because 1) not all exiftool commands need a file name and 2) Cmd can be used to provide a filename 
    ''' if it is given in the last Cmd before calling Execute</param>
    ''' <param name="MaxWait">Optional Integer: The maximum amount of time to wait for an Exiftool command to complete
    ''' execution. Defaults to 5 seconds (5000 milliseconds)</param>
    ''' <returns>String: the results of the exiftool command execution</returns>
    ''' <remarks>If the exiftool command execution returns a non-blank error msg that does not contain the word "warning"
    ''' then we will wait SleepStep (50) milliseconds and then excute the exiftool commands again.  It will repeat this
    ''' process until there is a no-error execution of the command or MaxWait time is past</remarks>
    Public Shared Function Execute(Optional fn As String = Nothing, Optional MaxWait As Integer = 5000) As String

        If Not InitializationDone Then Throw New Exception("ExifToolIO not initialized yet, can not use Execute")

        Dim CmdStr As String = CmdString.ToString

        CmdString.Clear()
        _LastErrorMsg = String.Empty

        Const SleepStep As Integer = 50

        Dim ntry As Integer = 0
        Dim so As String = ""
        Dim StartTime As Date = Now()

        Do
            If Not String.IsNullOrEmpty(CmdStr) Then CmdStream.Write(CmdStr)
            If Not String.IsNullOrEmpty(fn) Then CmdStream.WriteLine(fn)

            CmdStream.WriteLine(CmdExecute)

            If CmdStr.Contains(CmdQuiet & ControlChars.CrLf) OrElse CmdStr.EndsWith(CmdQuiet) OrElse
               CmdStr.Contains(CmdTableFormat & ControlChars.CrLf) OrElse CmdStr.EndsWith(CmdTableFormat) Then

                '--Do -execute again so we get the {ready} indicator 
                '--  (-q and -T options supress this and OutputDataHandler is waiting for it!)

                CmdStream.WriteLine(CmdExecute)
            End If

            '--wait for and get all the output from cmds sent to ExifTool

            so = stdout()

            '--did we just get a "warning" error message?  if so, then all is 'OK' and we are done

            If _LastErrorMsg <> String.Empty AndAlso Not _LastErrorMsg.ToLower.Contains("warning") Then
                If Past(StartTime, MaxWait) Then Throw New Exception("ExifTool execute failed after " & ntry & " attempts. Error: " & _LastErrorMsg)

                Thread.Sleep(SleepStep)
                ntry += 1
            Else
                If ntry > 0 Then

                    '--We eventually succeeded, but after more than one try, show that in the LastErrorMsg

                    If _LastErrorMsg <> String.Empty Then _LastErrorMsg &= ControlChars.CrLf
                    _LastErrorMsg &= "Warning: Execute took " & ntry \ SleepStep & " tries to succesfully execute"
                End If

                Exit Do
            End If
        Loop

        Return so
    End Function

    ''' <summary>
    ''' Return the last error/warning message from the last Execute
    ''' </summary>
    ''' <value></value>
    ''' <returns>error and/or warning message</returns>
    ''' <remarks></remarks>
    Public Shared ReadOnly Property LastErrorMsg() As String
        Get
            Return _LastErrorMsg
        End Get
    End Property

    ''' <summary>
    ''' This closes and cleans up the exiftool process started by Initialize.
    ''' </summary>
    ''' <param name="DeleteCmdFile">Boolean: Default is True, set to False to not delete the command file
    ''' that was used to send commands to exiftool.  This may prove useful for debugging exitool commands.
    ''' This file can then be looked at in a text editor to see what command where actually sent to exiftool.
    ''' For this to work, ExifToolIOOptions.UseStdInpCommandFile needs to be set to False and a cmd filename provided
    ''' in ExifToolIOOptions.CommandFileName</param>
    ''' <returns>Integer: This is the exit code returned by exitool.exe when the process is stopped, normally would be 0
    ''' See exiftool documentation for return error codes</returns>
    ''' <remarks>When the exiftool process starts there are acually 2 processes created </remarks>
    Public Shared Function Close(Optional DeleteCmdFile As Boolean = True) As Integer
        Dim ec As Integer = -1

        '--If the Initialization process is not complete yet, then wait for it to complerte so we can
        '--cleanly shut down.  Wait for up to 10 seconds

        Dim iwait As Integer = 0, istep As Integer = 100
        Do While Not InitializationDone
            Thread.Sleep(istep)
            iwait += istep
            If iwait > 10000 Then Exit Do
        Loop

        If InitializationDone Then
            SyncLock lockObject

                '--If created temporary ExifTool.exe file (because we needed to copy exiftool.exe from My.Resuources) 
                '--to run, then before closing, get the temporary files associated with this process

                Dim FileList As New List(Of String)
                If CreatedTempExe Then FileList = GetFilesAssociatedWithProcess(ExifToolProcess.ProcessName)

                '--close exiftool.exe

                CmdStream.WriteLine(CmdStayOpen)
                CmdStream.WriteLine("False")
                CmdStream.Close()

                '--Possibly delete the cmd file

                If CmdFn <> ExifToolIO.StdInpFileName AndAlso DeleteCmdFile Then DeleteFile(CmdFn)

                '--wait up to 10 seconds for Process to exit

                ExifToolProcess.WaitForExit(ProcessExitWaitTime)

                ec = ExifToolProcess.ExitCode

                RemoveHandler ExifToolProcess.OutputDataReceived, AddressOf OutputDataHandler
                RemoveHandler ExifToolProcess.ErrorDataReceived, AddressOf ErrorOutputHandler

                ExifToolProcess.Close()

                InitializationDone = False

                '--Delete the temporary Exiftool.exe file we created in Initialize

                For Each fn As String In FileList
                    DeleteFile(fn)
                Next
            End SyncLock
        End If

        Return ec
    End Function
#End Region

#Region "Private Subs and Functions"

    Private Shared Sub StartExifToolProcess()
        ExifToolProcess = New Process

        With ExifToolProcess
            .StartInfo.FileName = ExifToolExeFn_
            .StartInfo.UseShellExecute = False

            '--Use StandardInput or a file?

            If ExifToolCmdFn_ = ExifToolIO.StdInpFileName OrElse
               ExifToolCmdFn_.Equals("standardinput", StringComparison.OrdinalIgnoreCase) OrElse
               ExifToolCmdFn_.Equals("stdinp", StringComparison.OrdinalIgnoreCase) Then

                CmdFn = ExifToolIO.StdInpFileName
                .StartInfo.RedirectStandardInput = True
            Else
                If ExifToolCmdFn_ Is Nothing Then
                    CmdFn = Path.GetTempFileName
                Else
                    CmdFn = ExifToolCmdFn_
                End If

                CmdStream = New StreamWriter(CmdFn)
                CmdFn = """" & CmdFn & """"
            End If

            .StartInfo.Arguments = CmdStayOpen & " True -@ " & CmdFn

            .StartInfo.RedirectStandardOutput = True
            .StartInfo.RedirectStandardError = True
            .StartInfo.CreateNoWindow = True
            .StartInfo.WindowStyle = ProcessWindowStyle.Hidden

            .Start()

            If CmdFn = ExifToolIO.StdInpFileName Then CmdStream = .StandardInput
            CmdStream.AutoFlush = True  '--this is important!

            stdoutReadyHandle = New AutoResetEvent(False)

            AddHandler .OutputDataReceived, AddressOf OutputDataHandler
            .BeginOutputReadLine()

            erroutWaitHandle = New AutoResetEvent(False)

            AddHandler .ErrorDataReceived, AddressOf ErrorOutputHandler
            .BeginErrorReadLine()

        End With
    End Sub

    Private Shared Function Past(StartTime As Date, MaxWait As Integer) As Boolean
        Return (Now().Ticks - StartTime.Ticks) \ TimeSpan.TicksPerMillisecond > MaxWait
    End Function

    ''' <summary>
    ''' 
    ''' </summary>
    ''' <param name="pname"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Shared Function GetFilesAssociatedWithProcess(pname As String) As List(Of String)
        Dim FileList As New List(Of String)

        For Each p As Process In Process.GetProcesses
            If pname = p.ProcessName Then FileList.Add(p.MainModule.FileName)
        Next

        Return FileList
    End Function

    ''' <summary>
    ''' If a previous instance of this Application using ExifToolLib was closed abnormally or if ExifTool.Close was never called, then it is likely
    ''' the Process from it is still running.  This routine looks for Process' with an ExifTool type of name (8 digits and ExifToolProcessNameBase) 
    ''' and Kills them. If ExifExifTool was started from a .exe file that is not of this name format then this routine will not find it.
    ''' </summary>
    ''' <remarks></remarks>
    Private Shared Sub DoKillExifToolProcesses()
        Dim FilesToDelete As New List(Of String)
        Dim ProcessesToKill As New List(Of Process)

        '--First just make a list of all the ExifTool processes and associated file names

        For Each p As Process In Process.GetProcesses
            If (p.ProcessName.Length = (NTimeChars + ExifToolProcessNameBase.Length)) AndAlso
               (IsAllDigits(p.ProcessName.Substring(0, NTimeChars)) AndAlso p.ProcessName.EndsWith(ExifToolProcessNameBase)) Then

                FilesToDelete.Add(p.MainModule.FileName)
                ProcessesToKill.Add(p)
            End If
        Next

        '--Now Kill all the Processes we found

        For Each p As Process In ProcessesToKill
            Try
                p.Kill()

                '--wait up to 10 sec for the process to exit

                p.WaitForExit(ProcessExitWaitTime)

                p.Close()
            Catch winException As Win32Exception

                '--process was terminating or can't be terminated - deal with it 

            Catch invalidException As InvalidOperationException

                '--process has already exited - might be able to let this one go

            End Try
        Next

        '--Now delete all the associated files we found

        For Each fn As String In FilesToDelete
            DeleteFile(fn)
        Next
    End Sub

    Private Shared Sub OutputDataHandler(sender As Object, args As DataReceivedEventArgs)
        If args.Data IsNot Nothing Then

            '--handles each ReadLine

            If args.Data.ToLower.EndsWith(ReadyString) Then

                '--got the '{ready}' string. so we are done with output, so remove '{ready}' string and signal we are done

                stdoutString.Append(args.Data.Substring(0, args.Data.Length - ReadyString.Length))
                stdoutReadyHandle.Set()
            Else

                '--args.data does not have CRLf in it so we need to add that

                stdoutString.Append(args.Data & ControlChars.CrLf)
            End If
        End If
    End Sub

    Private Shared Function stdout() As String

        '--wait here until we get all the output which ends with a '{ready}'

        stdoutReadyHandle.WaitOne()

        Dim s As String = stdoutString.ToString
        stdoutString.Clear()
        _LastErrorMsg &= stderr()
        Return s
    End Function

    Private Shared Sub ErrorOutputHandler(sender As Object, args As DataReceivedEventArgs)
        If args.Data IsNot Nothing Then
            erroutString.Append(args.Data & ControlChars.CrLf)
            erroutWaitHandle.Set()
        End If
    End Sub

    Private Shared Function stderr() As String
        erroutWaitHandle.WaitOne(10)
        Dim s As String = erroutString.ToString
        erroutString.Clear()
        Return s
    End Function

    ''' <summary>
    ''' Deletes a file.  If not succesful immediately, then tries again each 100ms until 
    ''' MaxWait ms is reached, or file is succesfully deleted.
    ''' </summary>
    ''' <param name="fn"></param>
    ''' <param name="MaxWait"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Shared Function DeleteFile(fn As String, Optional MaxWait As Integer = 1500) As Boolean
        Const WaitStep As Integer = 100
        Dim ntry As Integer = 0
        Do
            Try
                File.Delete(fn)
                Return True
            Catch
                If ntry > MaxWait Then Return False

                Thread.Sleep(WaitStep)
                ntry += WaitStep
            End Try
        Loop
    End Function

    Private Shared Function IsAllDigits(str As String) As Boolean
        If String.IsNullOrEmpty(str) Then Return True

        For Each c As Char In str
            If Not (c >= "0"c AndAlso c <= "9"c) Then Return False
        Next

        Return True
    End Function
#End Region

End Class

